using System;
//using System.Collections.Generic;
using System.Text;
using System.IO;

#if NUNIT
using NUnit.Framework;
#endif
using Brecham.TestsInfrastructure;

using Brecham.Obex;
using InTheHand.Net.Sockets;
using InTheHand.Net;    //e.g. BluetoothAddress
using System.Net.Sockets;
using Brecham.Obex.Objects;   //e.g. NetworkStream


namespace GetFolderListings
{
    class Program
    {
        static void Main(string[] args)
        {
            //------------------------------
#if ! ( FX1_1 || NETCF )
            Console.Beep();
#endif
            // Check what protocol to connect over, we assume Bluetooth.
            Console.Write("Connect to a Bluetooth peer, select No for an IrDA peer");
            bool toBluetooth = MenuKeyboardInput.GetCmdLineIsYesAssumeYes();

            // Check what application to connect to, we assume Folder Browsing.
            Console.Write("Connect to the Folder-Browsing server, select No for the default inbox");
            bool fbService = MenuKeyboardInput.GetCmdLineIsYesAssumeYes();

            Console.WriteLine();
            
            //------------------------------
            using (TransportConnectionToObex conn = new TransportConnectionToObex(toBluetooth, fbService))
            {
                // Do it!
                cmdReader(conn.PeerStream, fbService);
                //
            }
        }

        
        static void cmdUsage()
        {
            Console.WriteLine("Commands:\n"
                + "Q    Quit.\n"
                + "\n"
                + "D    Dir the folder Listing.\n"
                + "L    Get the folder Listing.\n"
                + "\n"
                + "R    Reset to default folder.\n"
                + "U    Up one folder level.\n"
                + "C <name>   Change to folder with the given name.\n"
                + "\n"
                + "A    Get anonymous--no Name or Type.\n"
                + "T <type>   Get by Type alone.\n"
                + "N <name>    Get by Name.\n"
                + "B <name> <type>   Get by both Name and Type.\n"
                + "\n"
                + "e.g.\n"
                + "      cmd> c fooFolder\n"
                + "      cmd> g\n"
                + "      Downloading the folder listing.\n"
                + "      ... ...\n"
                + "      cmd> q\n"
                + "      Quitting.\n"
                );
        }//fn

        // Included in normal usage.
        //static void cmdUsageAdvanced()
        //{
        //    Console.WriteLine("Advanced Commands:\n"
        //        + "A    Get anonymous--no Name or Type.\n"
        //        + "T    <type>   Get by Type alone.\n"
        //        + "N    <name>    Get by Name.\n"
        //        + "B    <name> <type>   Get by both Name and Type.\n"
        //        );
        //}//fn

        
        static void cmdReader(Stream peer, bool fbService)
        {
            FolderListingMachinery machine;
            if (fbService)
            {
                Console.WriteLine("Connecting to Folder-Browsing Target.");
                machine = new FolderListingMachinery(peer);
            }
            else
            {
                Console.WriteLine("Connecting to null Target.");
                machine = new FolderListingMachinery(peer, null);
            }

            Console.WriteLine();
            cmdUsage();

            bool quit = false;
            while (!quit)
            {
                Console.Write("\ncmd>");
                String cmdLine = Console.ReadLine();
                if (cmdLine.Length == 0)
                    continue;
                Char cmdChar;
                if (cmdLine.Length == 1 || Char.IsWhiteSpace(cmdLine[1]))
                {
                    cmdChar = MenuKeyboardInput.FirstCharAsUpper(cmdLine);
                }
                else
                {
                    cmdChar = '\x0';
                }
                String cmdTail = cmdLine.Substring(1).Trim();
                try
                {
                    if ("hidden".Equals(cmdLine
#if ! FX1_1
                                    , StringComparison.OrdinalIgnoreCase
#endif
))
                    {
                        cmdUsage();
                        //cmdUsageAdvanced();
                    }
                    //----
                    else
                        if (cmdChar.Equals('Q'))    //Quit
                    {
                        Console.WriteLine("Quitting.");
                        quit = true;
                    }
                    //----
                    else if (cmdChar.Equals('D'))   //Dir the folder listing.
                    {
                        Console.WriteLine("Downloading the folder listing.");
                        machine.DirFolderListing();
                    } else if (cmdChar.Equals('L'))   //Get the folder listing.
                    {
                        Console.WriteLine("Downloading the folder listing.");
                        machine.GetFolderListing();
                    }
                        //----
                    else if (cmdChar.Equals('R'))   //Reset to default folder.
                    {
                        Console.WriteLine("Folder resetting to default.");
                        machine.FolderReset();
                    }
                    else if (cmdChar.Equals('U'))   //Change up one folder level.
                    {
                        Console.WriteLine("Folder change up one level.");
                        machine.FolderUp();
                    }
                    else if (cmdChar.Equals('C'))   //Change folder
                    {
                        Console.WriteLine("Folder change to {0}.", cmdTail);
                        machine.FolderChange(cmdTail);
                    }
                    //----
                    else if (cmdChar.Equals('A'))   //Get with no Name or Type.
                    {
                        Console.WriteLine("Get anonymous.");
                        Stream dest = machine.CreateFileStream("anonymous");
                        machine.ObexSession.GetTo(dest, null, null);
                    }
                    else if (cmdChar.Equals('T'))   //Get by Type alone.
                    {
                        Console.WriteLine("Get by TYPE={0}.", cmdTail);
                        Stream dest = machine.CreateFileStream("byType");
                        machine.ObexSession.GetTo(dest, null, cmdTail);
                    }
                    else if (cmdChar.Equals('N'))   //Get by Name.
                    {
                        Console.WriteLine("Get by NAME={0}.", cmdTail);
                        Stream dest = machine.CreateFileStream(
                            Path.GetFileName(cmdTail));
                        machine.ObexSession.GetTo(dest, cmdTail, null);
                    }
                    else if (cmdChar.Equals('B'))   //Get by both Name and Type.
                    {
                        int pos = cmdTail.LastIndexOf(' ');
                        String name = cmdTail.Substring(0, pos);
                        String type = cmdTail.Substring(pos + 1);
                        Console.WriteLine("Get by NAME={0}, TYPE={1}.", name, type);
                        Stream dest = machine.CreateFileStream(name + "_andType");
                        machine.ObexSession.GetTo(dest, name, type);
                    }
                    //----
                    else 
                    {
                        Console.WriteLine("Unrecognized command.");
                        cmdUsage();
                    }
                }//try
                catch (Exception ex)
                {
                    Type exType = ex.GetType();
                    Console.WriteLine("\n" + exType + ": " + ex.Message);
                    Console.Write("\nContinue after that exception");
                    if ( !MenuKeyboardInput.GetCmdLineIsYesAssumeNo())
                    {
                        throw;
                    }
                }//catch
                //
            }//while
        }
    }//class


    class TransportConnectionToObex : IDisposable
    {
        #region fields
        IrDAClient m_irdaCli;
        BluetoothClient m_btCli;
        NetworkStream m_peerStream;
        #endregion


        public TransportConnectionToObex(bool toBluetooth, bool fbService)
        {
            if (toBluetooth)
            {
                m_btCli = new BluetoothClient();
                BluetoothDeviceInfo[] devices;
                while (true)
                {
                    Console.WriteLine("Bluetooth discovery...");
                    devices = m_btCli.DiscoverDevices();
                    for (int i = 0; i < devices.Length; ++i)
                    {
                        BluetoothDeviceInfo curDev = devices[i];
                        Console.WriteLine("{0,2}: {1}, last seen: {2} [{3}]",
                            i, curDev.DeviceName, curDev.LastSeen, curDev);
                    }//for
#if ! ( FX1_1 || NETCF )
                    Console.Beep();
#endif
                    if (devices.Length != 0) break;
                    Console.Write("No devices discovered, hit return to try again>");
                    Console.ReadLine();
                }//while
                Console.Write("Choose which device by number");
                int selected = MenuKeyboardInput.GetIndexOfRange(devices.Length);
                BluetoothAddress addr = devices[selected].DeviceAddress;
                BluetoothEndPoint ep = new BluetoothEndPoint(addr,
                    fbService
                        ? InTheHand.Net.Bluetooth.BluetoothService.ObexFileTransfer
                        : InTheHand.Net.Bluetooth.BluetoothService.ObexObjectPush
                    );
                Console.WriteLine("Connecting to device...");
                m_btCli.Connect(ep);
                m_peerStream = m_btCli.GetStream();
            }
            else
            {
                m_irdaCli = new IrDAClient();
                IrDADeviceInfo[] devices;
                while (true)
                {
                    Console.WriteLine("IrDA discovery...");
                    devices = m_irdaCli.DiscoverDevices();
                    for (int i = 0; i < devices.Length; ++i)
                    {
                        IrDADeviceInfo curDev = devices[i];
                        Console.WriteLine("{0,2}: {1} [{3}]",//, last seen: {2}",
                            i, curDev.DeviceName, null//curDev.LastSeen
                            , curDev
                            );
                    }//for
#if ! ( FX1_1 || NETCF )
                    Console.Beep();
#endif
                    if (devices.Length != 0) break;
                    Console.Write("No devices discovered, hit return to try again>");
                    Console.ReadLine();
                }//while
                Console.Write("Choose which device by number");
                int selected = MenuKeyboardInput.GetIndexOfRange(devices.Length);
                IrDAAddress addr = devices[selected].DeviceAddress;
                IrDAEndPoint ep = new IrDAEndPoint(addr, "OBEX");
                Console.WriteLine("Connecting to device...");
                m_irdaCli = new IrDAClient(ep);
                m_peerStream = m_irdaCli.GetStream();
            }
        }//fn


        public void Close()
        {
            Dispose();
        }


        public NetworkStream PeerStream { get { return m_peerStream; } }


        #region IDisposable Members

        public void Dispose()
        {
            Dispose(true);
        }

        public void Dispose(bool disposing)
        {
            System.Diagnostics.Debug.Assert(disposing);
            if (disposing)
            {
                if (m_peerStream != null) m_peerStream.Close();
                if (m_irdaCli != null) m_irdaCli.Close();
                if (m_btCli != null) m_btCli.Close();
            }//if
        }

        #endregion
    }//class


#if ! FX1_1
    static 
#endif
    class MenuKeyboardInput
    {
        //----------------------------------------------------------------------
        public static int GetIndexOfRange(int range)
        {
            Int32 value = -99;
            while (true)
            {
                Console.Write(">");
                String line = Console.ReadLine();
                // No Int32.TryParse in CF20/FX1_1.
                // bool success = Int32.TryParse(line, out value);
                bool success = false;
                try
                {
                    //success = false;
                    value = Int32.Parse(line);
                    success = true;
                }
                catch (FormatException) { }
                catch (OverflowException) { }
                if (!success)
                {
                    Console.Write("Not a number, please try again");
                }
                else if (value >= range || value < 0)
                {
                    Console.Write("The number should be in the range, 0 through {0}", range - 1);
                }
                else
                { break; }
            }//while
            return value;
        }

        //----------------------------------------------------------------------
        public static bool GetCmdLineIsYesAssumeNo()
        {
            Console.Write(" [y/N]>");
            String input = Console.ReadLine();
            return GetCmdLineIsYesAssumeNo(input);
        }

        public static bool GetCmdLineIsYesAssumeYes()
        {
            Console.Write(" [Y/n]>");
            String input = Console.ReadLine();
            return GetCmdLineIsYesAssumeYes(input);
        }

        //----
        public static bool GetCmdLineIsYesAssumeNo(String input)
        {
            return GetCmdLineIsYesBlankIs(false, input);
        }

        public static bool GetCmdLineIsYesAssumeYes(String input)
        {
            return GetCmdLineIsYesBlankIs(true, input);
        }

        public static bool GetCmdLineIsYesBlankIs(bool assume, String input)
        {
            if( input == null
                || input.Length == 0)
                return assume;
            return FirstCharAsUpper(input) == 'Y';
        }

        //----------------------------------------------------------------------


        public static char FirstCharAsUpper(String s)
        {
#if ! FX1_1 && ! NETCF
            return s.Substring(0, 1).ToUpperInvariant()[0];
#else
            return s.Substring(0, 1).ToUpper(System.Globalization.CultureInfo.InvariantCulture)[0];
#endif
        }//fn

        //----------------------------------------------------------------------
    }//class



#if NUNIT
    [TestFixture]
    public class GetCmdLineIsYesEtc
    {
        //----------------------------------------
        [Test]
        public void noneAssumeFalse()
        {
            bool result = MenuKeyboardInput.GetCmdLineIsYesAssumeNo("");
            Assert.IsFalse(result);
        }

        [Test]
        public void noneAssumeTrue()
        {
            bool result = MenuKeyboardInput.GetCmdLineIsYesAssumeYes("");
            Assert.IsTrue(result);
        }

        //----------------------------------------
        [Test]
        public void noneBlankIsFalse()
        {
            bool result = MenuKeyboardInput.GetCmdLineIsYesBlankIs(false, "");
            Assert.IsFalse(result);
        }

        [Test]
        public void noneBlankIsTrue()
        {
            bool result = MenuKeyboardInput.GetCmdLineIsYesBlankIs(true, ""); 
            Assert.IsTrue(result);
        }

        //----------------------------------------
        [Test]
        public void nLower()
        {
            bool result = MenuKeyboardInput.GetCmdLineIsYesBlankIs(false, "n");
            Assert.IsFalse(result);
        }

        [Test]
        public void nUpper()
        {
            bool result = MenuKeyboardInput.GetCmdLineIsYesBlankIs(false, "N");
            Assert.IsFalse(result);
        }

        [Test]
        public void yLower()
        {
            bool result = MenuKeyboardInput.GetCmdLineIsYesBlankIs(false, "y");
            Assert.IsTrue(result);
        }

        [Test]
        public void yUpper()
        {
            bool result = MenuKeyboardInput.GetCmdLineIsYesBlankIs(false, "Y");
            Assert.IsTrue(result);
        }

        [Test]
        public void nWithMore()
        {
            bool result = MenuKeyboardInput.GetCmdLineIsYesBlankIs(false, "nKJSBDkasksaknakdhjd");
            Assert.IsFalse(result);
        }

        [Test]
        public void yWithMore()
        {
            bool result = MenuKeyboardInput.GetCmdLineIsYesBlankIs(false, "yjksai8ywqkhe12kj");
            Assert.IsTrue(result);
        }

    }//class
#endif


    internal class FolderListingMachinery
    {
        ObexClientSession m_obexSession;

        // The name of folder last changed to.
        // Except, if the last change was upward then its value is null.
        String m_lastFolderName = RootName;

        // The we assign to the default/base folder.
        public const String RootName = "Root";

        //-------------------

        internal FolderListingMachinery(Stream peerStream)
        {
            m_obexSession = new ObexClientSession(peerStream, UInt16.MaxValue);
            m_obexSession.Connect(ObexConstant.Target.FolderBrowsing);
        }

        internal FolderListingMachinery(Stream peerStream, byte[] target)
        {
            m_obexSession = new ObexClientSession(peerStream, UInt16.MaxValue);
            if (target != null)
            {
               m_obexSession.Connect(target);
            }
            else
            {
                m_obexSession.Connect();
            }
        }//fn

        //-------------------
        public ObexClientSession ObexSession { get { return m_obexSession; } }

        //-------------------

        public void FolderUp()
        {
            m_lastFolderName = null;
            m_obexSession.SetPathUp();
        }

        public void FolderReset()
        {
            m_lastFolderName = RootName;
            m_obexSession.SetPathUp();
        }

        public void FolderChange(String name)
        {
            m_lastFolderName = name;
            m_obexSession.SetPath(false, name, true);
        }

        public void FolderChangeAndMake(String name)
        {
            m_lastFolderName = name;
            m_obexSession.SetPath(name);
        }

        public void GetFolderListing()
        {
            /*
             * Spec section 8.1.1.1, item 1:
             * "To retrieve the current folder: Send a GET Request with an empty 
             * Name header and a Type header that specifies the folder object type."
             */
            Stream dest;
            if (m_lastFolderName == null)
            {
                Console.WriteLine("Just did an Up folder change, "
                    + "so don't know the current folder name, "
                    + "so not writing the folder-listing operation to disk...");
                dest = new TeeToConsoleStream(Stream.Null);
            }
            else
            {
                dest = CreateFileStream();
            }
            m_obexSession.GetTo(dest, null, ObexConstant.Type.FolderListing);
        }//fn


        DateTime adjustTimeForZoneDstEtc(DateTime dt)
        {
            return dt.AddHours(1);
        }

        //                       "19/07/2006  11:39            19,968 Backup of ToDo.wbk"
        //                       "29/07/2006  12:01    <DIR>          .."
        const String dirFormat = "{0,10:d}  {1,5:t}    {2,5} {3,8:n0} {4}";
        //                             "             305 File(s)     17,565,733 bytes";
        const String tailFilesFormat = "{0,16} File(s) {1,15:n0} bytes";
        //                            "              49 Dir(s)   2,236,100,608 bytes free"
        const String tailDirsFormat = "{0,16} Dir(s) {1,15:n0} bytes free";
        //
        public void DirFolderListing()
        {
            using (ObexGetStream strm = m_obexSession.Get(null, ObexConstant.Type.FolderListing)) {
                ObexFolderListingParser flp = new ObexFolderListingParser(strm);
                Brecham.Obex.Objects.ObexFolderListingItem item;
                //
                int countDirs = 0;
                int countFiles = 0;
                long totalSizeFiles = 0;
                //
                while ((item = flp.GetNextItem()) != null) {
                    if (item.GetType() == typeof(ObexParentFolderItem)) {
                        DateTime fakeDateTime = new DateTime();
                        // Include a completely fake "." directory when we see a parent-folder item.
                        Console.WriteLine(dirFormat, fakeDateTime, fakeDateTime, "<DIR>", null, ".");
                        countDirs++;
                        // The parent-folder item, sadly it doesn't have a 'modified' attribute.
                        Console.WriteLine(dirFormat, fakeDateTime, fakeDateTime, "<DIR>", null, "..");
                        countDirs++;
                    } else if (item.GetType() == typeof(ObexFileItem)) {
                        ObexFileOrFolderItem fofItem = (ObexFileItem)item;
                        DateTime dt = adjustTimeForZoneDstEtc(fofItem.Modified);
                        Console.WriteLine(dirFormat, dt, dt,
                            null, fofItem.Size, fofItem.Name);
                        countFiles++;
                        totalSizeFiles += fofItem.Size;
                    } else if (item.GetType() == typeof(ObexFolderItem)) {
                        ObexFileOrFolderItem fofItem = (ObexFolderItem)item;
                        DateTime dt = adjustTimeForZoneDstEtc(fofItem.Modified);
                        String size = null;
                        if (fofItem.Size != 0) { size = fofItem.Size.ToString(); }
                        Console.WriteLine(dirFormat, dt, dt,
                            "<DIR>", size, fofItem.Name);
                        countDirs++;
                    }
                }//while
                Console.WriteLine(tailFilesFormat, countFiles, totalSizeFiles);
                Console.WriteLine(tailDirsFormat, countDirs, -1);
            }//using
        }//fn

        //-------------------
        IMockFileStreamFactory m_fakeFileStreamFactory;
        public IMockFileStreamFactory FakeFileStreamFactory
        {
            get { return (IMockFileStreamFactory)m_fakeFileStreamFactory; }
            set { m_fakeFileStreamFactory = value; }
        }



        public Stream CreateFileStream()
        {
            String path;
            path = Path.ChangeExtension(m_lastFolderName, ".xml");
            return CreateFileStream(path);
        }

        public Stream CreateFileStream(String path)
        {
            if (m_fakeFileStreamFactory == null)
            {
                return new TeeToConsoleStream(
                    new FileStream(path, FileMode.Create));
            }
            else
            {
                return ((IMockFileStreamFactory)m_fakeFileStreamFactory)
                             .CreateFileStream(path);
            }
        }
        //public Stream CreateFileStream(Stream child)
        //{
        //    if (child == null) throw new ArgumentNullException("child");
        //    //
        //    return new TeeToConsoleStream(child);
        //}
    }//class

    internal class TeeToConsoleStream : Stream
    {
        Stream m_childStream;
        Decoder m_decoder;
        // Tee-ing to console is disabled, probably because previous write's
        // Decoder.Convert failed, likely because we're displaying a binary file.
        bool m_disabled;

        public TeeToConsoleStream(Stream destination)
        {
            m_childStream = destination;
        }//.ctor

#if ! FX1_1
        protected override void Dispose(bool disposing)
        {
#else
        public void Dispose()
        {
            bool disposing = true;
#endif
            try
            {
                if (!disposing || (this.m_childStream == null))
                {
                    // If finalizing, or we've already run...
                    return;
                }
                //else
                try
                {
                    this.Flush();
                }
                finally
                {
                    this.m_childStream.Close();
                }
            }
            finally
            {
                this.m_childStream = null;
#if ! FX1_1
                base.Dispose(disposing);
#endif
            }
        }//fn



        public override bool CanRead { get { return false; } }

        public override bool CanSeek { get { return false; } }

        public override bool CanWrite
        {
            get
            {
                if (this.m_childStream != null)
                {
                    return this.m_childStream.CanWrite;
                }
                return false;
            }
        }//prop

        public override void Flush()
        {
            if (this.m_childStream == null)
            {
                throwIsDisposed();
            }

            // Note: A Decoder surely can only produce a character when all its 
            // bytes have arrived.  Flushing it just resets it  telling it 
            // that any subsequent bytes are of a new sequence.

            m_childStream.Flush();
        }//fn

        private void throwIsDisposed()
        {
            const String ExMsg_IsClosed = "ExMsg_IsClosed";
            throw new ObjectDisposedException(ExMsg_IsClosed);
        }

        public override long Length
        {
            get { throw new Exception("The method or operation is not implemented."); }
        }

        public override long Position
        {
            get { throw new Exception("The method or operation is not implemented."); }
            set { throw new Exception("The method or operation is not implemented."); }
        }

        public override int Read(byte[] buffer, int offset, int count)
        {
            throw new Exception("The method or operation is not implemented.");
        }

        public override long Seek(long offset, SeekOrigin origin)
        {
            throw new Exception("The method or operation is not implemented.");
        }

        public override void SetLength(long value)
        {
            throw new Exception("The method or operation is not implemented.");
        }

        public override void Write(byte[] buffer, int offset, int count)
        {
            if (buffer == null)
            {
                throw new ArgumentNullException("buffer");
            }
            if (offset < 0)
            {
                throw new ArgumentOutOfRangeException("offset");
            }
            if (count < 0)
            {
                throw new ArgumentOutOfRangeException("count");
            }
            if ((buffer.Length - offset) < count)
            {
                throw new ArgumentException("Argument_InvalidOffLen");
            }
            //
            if (this.m_childStream == null)
            {
                throwIsDisposed();
            }
            
            // Write to console
            if (!m_disabled) {
                WriteToConsole(buffer, offset, count);
            }else{
                Console.Write(".");
            }
            // Now tee to child
            m_childStream.Write(buffer, offset, count);
        }//fn


        bool IsPrefix(byte[] expected, byte[] buffer, int offset, int count)
        {
            if (offset + count > buffer.Length)
                throw new ArgumentOutOfRangeException();
            if (count < expected.Length)
                throw new ArgumentOutOfRangeException();
            for (int i = 0; i < expected.Length; ++i)
            {
                if (expected[i] != buffer[offset + i])
                    return false;
            }
            return true;
        }//

        public char[] Convert(byte[] buffer, int offset, int count,
            out int countOfChars)
        {
            if (m_decoder == null)
            {
                const int NumBytesToDoDetection=3;
                if (count < NumBytesToDoDetection)
                    throw new ArgumentOutOfRangeException("count",
                        "Need the first Write to be at least " 
                        + NumBytesToDoDetection 
                        + " bytes to allow encoding detection.");
                Encoding encg;
                if (buffer[0] == 0
                    || IsPrefix(Encoding.BigEndianUnicode.GetPreamble(), buffer, offset, count))
                {
                    encg = Encoding.BigEndianUnicode;
                }//if
                else if (buffer[1] == 0
                        || IsPrefix(Encoding.Unicode.GetPreamble(), buffer, offset, count))
                {
                    encg = Encoding.Unicode;
                }//else
                else
                {
                    encg = Encoding.UTF8;
                }//else
                //
                m_decoder = encg.GetDecoder();
#if ! FX1_1
                m_decoder.Reset();
#endif
            }//if
            char[] chars = new char[1024];
            int numChars = m_decoder.GetChars(buffer, offset, count, chars, 0);
            countOfChars = numChars;
            return chars; 
        }

        public void WriteToConsole(byte[] buffer, int offset, int count)
        {
            int cch;
            try {
                char[] chars = Convert(buffer, offset, count, out cch);
                Console.Write(chars, 0, cch);
            } catch(Exception ex) {
                m_disabled = true;
                Console.WriteLine("Error while displaying, disabling displaying. ["
                    + ex.GetType() + " " + ex.Message + "]");
            } 
#if FX1_1
            // In FX2 non-Exception-derived exceptions are wrapped in a 
            // System.Runtime.CompilerServices.RuntimeWrappedException.
            // But they aren't in FX1_1, so handle them here.
            catch {
                m_disabled = true;
                Console.WriteLine("Error while displaying, disabling displaying.");
            }
#endif
        }//fn

    }//class

#if NUNIT
    [TestFixture]
    public class Test_TeeStream
    {
        public static void Assert_AreEqual_CharBuffer(char[] expected, char[] result, int filledLength)
        {
            char[] resultComplete;
            // If the result array is only partially filled resize it.
            if (result.Length != filledLength)
            {
                resultComplete = new char[filledLength];
                Array.Copy(result, resultComplete, filledLength);
            }
            else
            {
                resultComplete = result;
            }
            Assert.AreEqual(expected, resultComplete);
        }//fn

        //------------------------------------------------
        void doTest(char[] expected, byte[] input)
        {
            TeeToConsoleStream strm = new TeeToConsoleStream(Stream.Null);
            doTest(strm, expected, input);
        }
        void doTest(TeeToConsoleStream strm, char[] expected, byte[] input)
        {
            int cch;
            char[] chars = strm.Convert(input, 0, input.Length, out cch);
            Assert.AreEqual(expected.Length, cch);
            Assert_AreEqual_CharBuffer(expected, chars, cch);
        }

        //------------------------------------------------
        [Test]
        public void initUtf8Noprefix()
        {
            byte[] input = { 0x61, 0x62, 0x63, 0x64 };
            char[] expected = new char[] { 'a', 'b', 'c', 'd' };
            doTest(expected,input);
        }

        [Test]
        public void initUtf8Withprefix()
        {
            byte[] input = { 0xEF, 0xBB, 0xBF, 
                0x61, 0x62, 0x63, 0x64 };
            char[] expected = new char[] { '\uFEFF', 'a', 'b', 'c', 'd' };
        }

        [Test]
        public void threeUtf8()
        {
            TeeToConsoleStream strm = new TeeToConsoleStream(Stream.Null);
            byte[] input = { 0x61, 0x62, 0x63, 0x64 };
            char[] expected = new char[] { 'a', 'b', 'c', 'd' };
            doTest(strm, expected, input);
            byte[] input1 = { 0x7A, 0xC3,0xB6, 0x30, 0x44 };
            char[] expected1 = new char[] { 'z', '\u00F6', '0', 'D' };
            doTest(strm, expected1, input1);
            byte[] input2 = { 0x20, 0x62, 0x63, 0x64 };
            char[] expected2 = new char[] { ' ', 'b', 'c', 'd' };
            doTest(strm, expected2, input2);
        }

        [Test]
        public void threeUtf8Split()
        {
            TeeToConsoleStream strm = new TeeToConsoleStream(Stream.Null);
            byte[] input = { 0x61, 0x62, 0x63, 0x64 };
            char[] expected = new char[] { 'a', 'b', 'c', 'd' };
            doTest(strm, expected, input);
            byte[] input1 = { 0x7A, 0xC3,};
            char[] expected1 = new char[] { 'z', };
            doTest(strm, expected1, input1);
            byte[] input2 = { 0xB6, 0x30, 0x44 };
            char[] expected2 = new char[] { '\u00F6', '0', 'D' };
            doTest(strm, expected2, input2);
        }

        [Test]
        public void initBigeNoprefix()
        {
            byte[] input = { 0,0x61, 0,0x62, 0,0x63, 0,0x64, 0x20,0x30 };
            char[] expected = new char[] { 'a', 'b', 'c', 'd', '\u2030' };
            doTest(expected, input);
        }

        [Test]
        public void initBigeWithprefix()
        {
            byte[] input = { 0xFE, 0xFF, 
                0,0x61, 0,0x62, 0,0x63, 0,0x64, 0x20,0x30 };
            char[] expected = new char[] { '\uFEFF', 'a', 'b', 'c', 'd', '\u2030' };
        }
        [Test]
        public void initLitleNoprefix()
        {
            byte[] input = { 0x61,0, 0x62,0, 0x63,0, 0x64,0, 0x30,0x20};
            char[] expected = new char[] { 'a', 'b', 'c', 'd', '\u2030' };
            doTest(expected, input);
        }

        [Test]
        public void initLitleWithprefix()
        {
            byte[] input = { 0xFF, 0xFE, 
                0x61,0, 0x62,0, 0x63,0, 0x64,0, 0x30,0x20, };
            char[] expected = new char[] { '\uFEFF', 'a', 'b', 'c', 'd', '\u2030' };
        }

        [Test]
        public void litleWithprefixSplit()
        {
            byte[] input = { 0xFF, 0xFE, 
                0x61,0, 0x62,0, 0x63,};
            byte[] input1 = { 0, 0x64,0, 0x30,0x20, };
            char[] expected = new char[] { '\uFEFF', 'a', 'b', };
            char[] expected1 =             { 'c', 'd', '\u2030' };
            TeeToConsoleStream strm = new TeeToConsoleStream(Stream.Null);
            doTest(strm, expected, input);
            doTest(strm, expected1, input1);
        }

        [Test]
        public void bigeWithprefixSplit()
        {
            byte[] input = { 0xFE,0xFF,
                0,0x61, 0,0x62,0,};
            byte[] input1 = { 0x63, 0,0x64, 0x20,0x30 };
            char[] expected = new char[] { '\uFEFF', 'a', 'b', };
            char[] expected1 =             { 'c', 'd', '\u2030' };
            TeeToConsoleStream strm = new TeeToConsoleStream(Stream.Null);
            doTest(strm, expected, input);
            doTest(strm, expected1, input1);
        }


        [Test]
        public void litleWithSuppPairWithprefixSplit()
        {
            // Also splits the codepoint U+10000, which is the surrogate pair,
            // U+D800, U+DC00, over two byte arrays.
            byte[] input = { 0xFF, 0xFE, 0x61, 0, 0x62, 0, 0x63, };
            byte[] input1 = { 0, 0x64, 0, 0x30, 0x20, 
                /*supp. pair starts */0x00, 0xD8, 0x00 };
            byte[] input2 = { 0xDC };
            char[] expected = new char[] { '\uFEFF', 'a', 'b', };
            char[] expected1 = { 'c', 'd', '\u2030' };
            char[] expected2 = "\U00010000".ToCharArray();
            TeeToConsoleStream strm = new TeeToConsoleStream(Stream.Null);
            doTest(strm, expected, input);
            doTest(strm, expected1, input1);
            doTest(strm, expected2, input2);
        }

    }//class


    //[TestFixture]
    public class Test_Ops
    {

        [Test]
        /*public*/ void testAFew() 
        {
            MemoryStream bodyStream = new MemoryStream();
            MemoryStream rcvStrm = new MemoryStream(64);
            //
            const int LengthGet1 = 10;
            const int LengthGet2 = 9;
            const int LengthGet3 = 11;
            byte[] response ={ 
                //--auto Connect--
                0xA0,0,26, /**/0x10,0,0xFF,0xFF,
                    // Who
                        0x4A,0,19, 0xF9, 0xEC, 0x7B, 0xC4, 0x95, 0x3C, 0x11, 
                        0xd2, 0x98, 0x4E, 0x52, 0x54, 0x00, 0xDC, 0x9E, 0x09,
                //---- ----
                /*continue*/0x90,0,16,0x48,0,13,1,2,3,4,5,6,7,8,9,10,
                /*ok*/0xA0, 0, 6, 0x49,0,3,
                //--SetPath--
                0xA0,0,3,
                /*continue*/0x90,0,15,0x48,0,12,101,2,3,4,5,6,7,8,9,
                /*ok*/0xA0, 0, 6, 0x49,0,3,
                //--SetPath--
                0xA0,0,3,
                //--SetPath--
                0xA0,0,3,
                /*continue*/0x90,0,17,0x48,0,14,201,2,3,4,5,6,7,8,9,10,99,
                /*ok*/0xA0, 0, 6, 0x49,0,3,
            };
            MemoryStream rspStrm = new MemoryStream(response, false);
            PairOfStream strm = new PairOfStream(rspStrm, rcvStrm);
            //
            IMockFileStreamFactory fileStreamFactory = new MyMockFileStreamFactory();
            //
            FolderListingMachinery machine = new FolderListingMachinery(strm);
            machine.FakeFileStreamFactory = fileStreamFactory;
            String f1 = "f1";
            String f2 = "f2";
            //
            machine.GetFolderListing();
            Assert.AreEqual(new String[] { FolderListingMachinery.RootName },
                fileStreamFactory.GetGivenPaths());
            Assert.AreEqual(LengthGet1, 
                ((MockFileStream)fileStreamFactory.CurrentStream).WrittenLength);
            machine.FolderChange(f1);
            machine.GetFolderListing();
            Assert.AreEqual(LengthGet2,
                ((MockFileStream)fileStreamFactory.CurrentStream).WrittenLength);
            machine.FolderUp();
            machine.FolderChange(f2);
            machine.GetFolderListing();
            Assert.AreEqual(LengthGet3,
                ((MockFileStream)fileStreamFactory.CurrentStream).WrittenLength);
            Assert.AreEqual(new String[] { FolderListingMachinery.RootName, f1, f2 },
                fileStreamFactory.GetGivenPaths());
        }

        [Test]
        //[ExpectedException]
        /*public*/ void testGetAfterUp()
        {
            MemoryStream bodyStream = new MemoryStream();
            MemoryStream rcvStrm = new MemoryStream(64);
            //
            byte[] response ={ 
                //--auto Connect--
                0xA0,0,26, /**/0x10,0,0xFF,0xFF,
                    // Who
                        0x4A,0,19, 0xF9, 0xEC, 0x7B, 0xC4, 0x95, 0x3C, 0x11, 
                        0xd2, 0x98, 0x4E, 0x52, 0x54, 0x00, 0xDC, 0x9E, 0x09,
                //--SetPath--
                0xA0,0,3,
                //--SetPath--
                0xA0,0,3,
                /*continue*/0x90,0,16,0x48,0,13,201,2,3,4,5,6,7,8,9,10,
                /*ok*/0xA0, 0, 6, 0x49,0,3,
            };
            MemoryStream rspStrm = new MemoryStream(response, false);
            PairOfStream strm = new PairOfStream(rspStrm, rcvStrm);
            //
            IMockFileStreamFactory fileStreamFactory = new MyMockFileStreamFactory();
            //
            FolderListingMachinery machine = new FolderListingMachinery(strm);
            machine.FakeFileStreamFactory = fileStreamFactory;
            String f1 = "f1";
            //
            machine.FolderChange(f1);
            machine.FolderUp();
            machine.GetFolderListing();
        }

        [Test]
        /*public*/ void testTwoDeep() 
        {
            MemoryStream bodyStream = new MemoryStream();
            MemoryStream rcvStrm = new MemoryStream(64);
            //
            byte[] response ={ 
                //--auto Connect--
                0xA0,0,26, /**/0x10,0,0xFF,0xFF,
                    // Who
                        0x4A,0,19, 0xF9, 0xEC, 0x7B, 0xC4, 0x95, 0x3C, 0x11, 
                        0xd2, 0x98, 0x4E, 0x52, 0x54, 0x00, 0xDC, 0x9E, 0x09,
                //--SetPath--
                0xA0,0,3,
                //--SetPath--
                0xA0,0,3,
                /*continue*/0x90,0,16,0x48,0,13,201,2,3,4,5,6,7,8,9,10,
                /*ok*/0xA0, 0, 6, 0x49,0,3,
            };
            MemoryStream rspStrm = new MemoryStream(response, false);
            PairOfStream strm = new PairOfStream(rspStrm, rcvStrm);
            //
            IMockFileStreamFactory fileStreamFactory = new MyMockFileStreamFactory();
            //
            FolderListingMachinery machine = new FolderListingMachinery(strm);
            machine.FakeFileStreamFactory = fileStreamFactory;
            String f1 = "f1";
            String f1b = "f1b";
            //
            machine.FolderChange(f1); 
            machine.FolderChange(f1b);
            machine.GetFolderListing();
            Assert.AreEqual(new String[] { f1b },
                fileStreamFactory.GetGivenPaths());
        }

    }//class
#endif

    class MyMockFileStreamFactory : IMockFileStreamFactory
    {
        public MyMockFileStreamFactory()
        {
            //m_paths = new List/*<string>*/();
            m_paths = new System.Collections.ArrayList/*<string>*/();
        }
        public Stream CreateFileStream(String path)
        {
            Stream stream = new MockFileStream(path);
            m_lastCreatedStream = stream;
            AppendPath(path);
            return stream;
        }

        public Stream CurrentStream { get { return m_lastCreatedStream;} }

        public void AppendPath(String path)
        {
            m_paths.Add(path);
        }

        public string[] GetGivenPaths()
        {
            String[] xxx = (String[])m_paths.ToArray();
            return xxx;
        }

        System.Collections.ArrayList/*<String>*/ m_paths;
        Stream m_lastCreatedStream;
    }

}//namespace
