using System;
//using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
//
using Brecham.Obex;
//
using InTheHand.Net.Sockets;
using InTheHand.Net;
using System.Net.Sockets;   //e.g. NetworkStream


namespace PutGUI
{


    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
            this.labelStatus.Text = "";
        }

        private void button1_Click(object sender, EventArgs e)
        {
            DialogResult result = openFileDialog1.ShowDialog();
            if (result == DialogResult.OK)
            {
                State state = new State();

                //------------------------------------------------------
                // Get the file
                //------------------------------------------------------
                String putName; // = "dummy.txt";
                state.m_fileStream = new FileStream(openFileDialog1.FileName, FileMode.Open, FileAccess.Read);
                state.m_progressStream = new ReadProgressStream(state.m_fileStream);
                state.m_progressStream.SetTotalReadLength(state.m_fileStream.Length);
                putName = Path.GetFileName(openFileDialog1.FileName);

                //------------------------------------------------------
                // Get the peer
                //------------------------------------------------------
                this.labelStatus.Text = "Connecting...";
                bool toBluetooth;
                if (rbProtoIrda.Checked) toBluetooth = false;
                else if (rbProtoBluetooth.Checked) toBluetooth = true;
                else
                {
                    System.Diagnostics.Debug.Fail("Unknown protocols radio box selected.");
                    throw new InvalidOperationException("Unknown protocols radio box selected.");
                }
                state.m_conn = new Brecham.Obex.Net.GuiObexSessionConnection(toBluetooth, false, this.labelStatus);
                if (!state.m_conn.Connect()) {
                    //user cancelled the connect
                    return;
                }
                Stream peerStream = state.m_conn.PeerStream;

                //------------------------------------------------------
                // Send
                //------------------------------------------------------
                try
                {
                    ObexClientSession sess = state.m_conn.ObexClientSession;
                    //
                    this.labelStatus.Text = "Sending...";
                    this.progressBar1.Visible = true;
                    StartProgressBarUpdater(state);
                    //sess.PutFrom(state.m_progressStream, putName, null, state.m_fileStream.Length);
                    state.m_putCaller = new PutFromNtiCaller(sess.PutFrom);
                    AsyncCallback cb = new AsyncCallback(PutCompleted);
                    state.SetStartTime();
                    IAsyncResult ar = state.m_putCaller.BeginInvoke(
                        state.m_progressStream, putName, null, state.m_fileStream.Length,
                        cb, state);
                }
                catch
                {
                    // Just ensure the streams are closed etc, and rethrow.
                    state.Dispose();
                    throw;
                }
            }//if
        }

        delegate void PutFromNtiCaller(Stream source, String name, String type, Int64 length);


        //----------------------------------------------------------------------
        class State : IDisposable
        {
            public Stream m_fileStream;
            public ReadProgressStream m_progressStream;
            public Brecham.Obex.Net.ObexSessionConnection m_conn;
            public PutFromNtiCaller m_putCaller;
            //
            private DateTime m_dtStart;
            private DateTime m_dtEnd;


            public void SetStartTime()
            {
                m_dtStart = DateTime.UtcNow;
            }

            /// <summary>
            /// Set the end time.  If called more than once, only the first is effective.
            /// </summary>
            public void SetEndTime()
            {
                if (m_dtEnd.Ticks == 0)
                {
                    m_dtEnd = DateTime.UtcNow;
                }//if
            }

            public TimeSpan Elapsed
            {
                get
                {
                    if (m_dtEnd.Ticks==0 || m_dtStart.Ticks==0)
                    {
                        throw new InvalidOperationException("Start or end time not set.");
                    }
                    TimeSpan ts = m_dtEnd - m_dtStart;
                    if (ts < TimeSpan.Zero)
                    {
                        throw new InvalidOperationException("End time is before start time.");
                    }
                    return ts;
                }
            }

            #region IDisposable Members

            public void Dispose()
            {
                if (m_conn != null) m_conn.Close();
                if (m_progressStream != null) m_progressStream.Close();
                if (m_fileStream != null) m_fileStream.Close();
            }

            #endregion
        }


        //----------------------------------------------------------------------
        // There's no way of passing data to the Timer's event hendler so we
        // unfortunately need this 'global' reference.  In desktop FXv2 there's a 
        // Tag property which might do the job but we won't use it for compatibility 
        // reasons.
        ReadProgressStream m_m_progressStreamReferenceForTimerBooHoo;

        private void timer1_Tick(object sender, EventArgs e)
        {
            this.progressBar1.Value = (int)m_m_progressStreamReferenceForTimerBooHoo.ReadPercentage;
        }

        private void StartProgressBarUpdater(State state)
        {
            m_m_progressStreamReferenceForTimerBooHoo = state.m_progressStream;
            this.timer1.Enabled = true;
        }//fn--button1_Click

        private void EndProgressBarUpdater()
        {
            this.timer1.Enabled = false;
            // just for tidyness.
            m_m_progressStreamReferenceForTimerBooHoo = null;
        }

        //----------------------------------------------------------------------
        void PutCompleted(IAsyncResult ar)
        {
            State state = (State)ar.AsyncState;
            state.SetEndTime(); //this only has an effect on the first call

            // We're going to update the UI so ensure that we're running on the UI 
            // thread, if necessary by calling ourselves back with Control.Invoke.
            if (this.InvokeRequired)
            {
                System.Diagnostics.Debug.WriteLine("InvokeRequired.");
                // (Argument '2' is not a params array in FX1.1)
                this.Invoke(new AsyncCallback(this.PutCompleted), new object[] { ar });
                return;
            }

            // Now do the work...
            try
            {
                EndProgressBarUpdater();
                try
                {
                    // Get the result of the Put operation. This doesn't need to 
                    // be on the UI thread, but the rest does...
                    state.m_putCaller.EndInvoke(ar);
                    this.labelStatus.Text = "PutFrom took: " + state.Elapsed.ToString();
                }
                finally
                {
                    state.Dispose();
                }
            }
            catch (Exception ex)
            {
                // Intending to to catch exceptions from the Put operation...
                // Feel ok to catch all here, as anything catastrophic would 
                // surely have occured on the UI thread too...
                String descr = ex.GetType().ToString() + "\r\n" + ex.Message;
                this.labelStatus.Text = "PutFrom failed: " + descr;
                MessageBox.Show(
#if ! NETCF
                    //The IWin32Window argument is not supported by .NETCF.  
                    this, 
#endif // ! NETCF
                    descr, "PutFrom failed");
            }
        }


    }

}//namespace

