
Option Explicit
Option Strict

Imports System
Imports System.Windows.Forms
Imports System.ComponentModel   'e.g. BackgroundWorker
Imports System.Drawing  ' e.g. Point
Imports System.IO
Imports InTheHand.Net.Sockets
Imports Brecham.Obex
Imports Brecham.Obex.Net

Class PutGuiVb
    Inherits Form

    '---------------------------------------------------------------------------
    '---------------------------------------------------------------------------
    Private WithEvents button1 As Button
    Private WithEvents buttonCancel As Button
    Private label1 As Label
    Private openFileDialog1 As OpenFileDialog
    Private WithEvents backgroundWorker1 As BackgroundWorker
    Private progressBar1 As ProgressBar
    Private rbProtoBluetooth As RadioButton
    Private rbProtoIrda As RadioButton
      
     
    '---------------------------------------------------------------------------
    '---------------------------------------------------------------------------
    Public Sub New
        button1 = New Button
        button1.Text = "B&rowse..."
        button1.Location = New Point(150, 150)
        '
        buttonCancel = New Button
        buttonCancel.Text = "&Cancel"
        buttonCancel.Location = New Point(150, 210)
        buttonCancel.Enabled = False
        '
        label1 = New Label
        label1.Location = New Point(10, 130)
        label1.Size = New Size(250, 15)
        label1.Text = "To choose the file to send, click Browse."
        '
        openFileDialog1 = New OpenFileDialog
        '
        backgroundWorker1 = New BackgroundWorker
        backgroundWorker1.WorkerReportsProgress = True
        backgroundWorker1.WorkerSupportsCancellation = True
        '
        progressBar1 = New ProgressBar
        progressBar1.Anchor = AnchorStyles.Right Or AnchorStyles.Left
        progressBar1.Location = New Point(10, 180)
        progressBar1.Size = New Size(Me.ClientSize.Width-20, 23)
        '
        rbProtoIrda = New RadioButton
        rbProtoIrda.Checked = True
        rbProtoIrda.Location = New Point(13, 13)
        rbProtoIrda.Name = "rbProtoIrda"
        rbProtoIrda.Size = New Size(46, 17)
        rbProtoIrda.TabIndex = 2
        rbProtoIrda.TabStop = True
        rbProtoIrda.Text = "&IrDA"
        rbProtoBluetooth = New RadioButton
        rbProtoBluetooth.Location = New Point(13, 37)
        rbProtoBluetooth.Name = "rbProtoBluetooth"
        rbProtoBluetooth.Size = New Size(70, 17)
        rbProtoBluetooth.TabIndex = 3
        rbProtoBluetooth.Text = "&Bluetooth"
        '
        Me.Text = "PutGuiVb"
        '
        Controls.Add(button1)
        Controls.Add(buttonCancel)
        Controls.Add(label1)
        Controls.Add(progressBar1)
        Controls.Add(rbProtoIrda)
        Controls.Add(rbProtoBluetooth)
    End Sub
    
    '---------------------------------------------------------------------------
    '---------------------------------------------------------------------------
    Private Sub button1_Click(ByVal sender As Object, ByVal e As EventArgs) Handles button1.Click
        label1.Text = "Choose a file."
        Dim result1 As DialogResult = openFileDialog1.ShowDialog(Me)
        If (result1 = System.Windows.Forms.DialogResult.OK) Then
            ' Open the source file
            label1.Text = "Opening the source file..."
            Dim src As New FileStream(openFileDialog1.FileName, FileMode.Open, FileAccess.Read)
            ' Connect to the peer device and service
            label1.Text = "Connecting..."
            Dim toBluetooth As Boolean
            If (rbProtoIrda.Checked) Then
                toBluetooth = False
            ElseIf (rbProtoBluetooth.Checked) Then
                toBluetooth = True
            Else
                System.Diagnostics.Debug.Fail("unknown protocol radiobutton checked")
                Exit Sub
            End If
            Dim conn As New GuiObexSessionConnection(toBluetooth, false, label1)
            Dim connected As Boolean = conn.Connect
            If (Not connected) Then Exit Sub
            ' Initiate the PUT now
            label1.Text = "Initiating the PUT..."
            Dim filename As String = Path.GetFileName(openFileDialog1.FileName)
            Dim putStream As ObexPutStream = conn.ObexClientSession.Put(filename, Nothing)
            buttonCancel.Enabled = True
            backgroundWorker1.RunWorkerAsync(New BgWorkerArgs(putStream, src, conn))
        End If
    End Sub


    Private Sub buttonCancel_Click(ByVal sender As Object, ByVal e As EventArgs) Handles buttonCancel.Click
        backgroundWorker1.CancelAsync
    End Sub
    
    
    '---------------------------------------------------------------------------
    '---------------------------------------------------------------------------
    Sub DoWork(ByVal sender As Object, ByVal e As DoWorkEventArgs) _
    Handles backgroundWorker1.DoWork
        Dim args As BgWorkerArgs = CType(e.Argument, BgWorkerArgs)
        Dim buffer(1023) As Byte
        Dim length As Int32
        Dim updatePeriod As New TimeSpan(0,0,0,0, 250)
        Dim lastProgress As DateTime = DateTime.UtcNow
        Dim percentage As Int32
        Dim elapsed As TimeSpan
        label1.Text = "Sending PUT content..."
        Try
            While(True)
                If (backgroundWorker1.CancellationPending)
                    e.Cancel = True
                    args.PutStream.Abort("User cancelled")
                    Exit While
                End If
                length = args.Source.Read(buffer, 0, buffer.Length)
                If(length = 0)
                    Exit While
                End If
                args.putStream.Write(buffer, 0, length)
                ' Progress reporting; rate-limited.
                elapsed = DateTime.UtcNow - lastProgress
                If (elapsed > updatePeriod) Then
                    'Console.Write("(" +args.Source.Position.ToString +"/"+args.Source.Length.ToString+ "=" + ((100.0 * args.Source.Position) / args.Source.Length).ToString +") ") 
                    percentage = CType((100.0 * args.Source.Position) / args.Source.Length,Int32)
                    backgroundWorker1.ReportProgress(percentage)
                    lastProgress = DateTime.UtcNow
                End If
            End While
            ' The final update of the progress bar
            percentage = CType((100.0 * args.Source.Position) / args.Source.Length,Int32)
            backgroundWorker1.ReportProgress(percentage)
        Finally
            args.PutStream.Close
            args.Source.Close
            ' If we wanted to do another Put to the same peer we wouldn't close 
            ' the session and its underlying network stream here.
            args.Connection.ObexClientSession.Dispose
        End Try
    End Sub

	' Update the progress bar on the UI thread
    Sub ProgressChanged(ByVal sender As Object, ByVal e As ProgressChangedEventArgs) Handles backgroundWorker1.ProgressChanged
        'Console.Write("" + e.ProgressPercentage.ToString + "% ")
        progressBar1.Value = e.ProgressPercentage
    End Sub

    Sub RunWorkerCompleted(ByVal sender As Object, ByVal e As RunWorkerCompletedEventArgs) Handles backgroundWorker1.RunWorkerCompleted
        buttonCancel.Enabled = False
        If (e.Cancelled) Then
            label1.Text = "Cancelled"
            MessageBox.Show("Cancelled")
        Else
            ' Propagate any exception that occured in DoWork!
            Dim obj As Object = e.Result
            label1.Text = "Complete"
            MessageBox.Show("Complete")
        End If
    End Sub

    '---------------------------------------------------------------------------
    '---------------------------------------------------------------------------
    Class BgWorkerArgs
        ' Fields
        Private ReadOnly m_putStream As ObexPutStream
        Private ReadOnly m_source As Stream
        Private ReadOnly m_connection As ObexSessionConnection
        
        ' Properties
        Public ReadOnly Property PutStream As ObexPutStream
        Get
            Return m_putStream
        End Get
        End Property
        
        Public ReadOnly Property Source As Stream
        Get
            Return m_source
        End Get
        End Property
        
        Public ReadOnly Property Connection As ObexSessionConnection
        Get
            Return m_connection
        End Get
        End Property

       
        ' Constructor
        Public Sub New(putStream As ObexPutStream, source As Stream, conn As ObexSessionConnection)
            If IsNothing(putStream) Then Throw New ArgumentNullException("putStream")
            If IsNothing(source) Then Throw New ArgumentNullException("source")
            If IsNothing(conn) Then Throw New ArgumentNullException("conn")
            m_putStream = putStream
            m_source = source
            m_connection = conn
        End Sub
    End Class


    '---------------------------------------------------------------------------
    '---------------------------------------------------------------------------
    Public Shared Sub Main
        Application.Run(New PutGuiVb)
    End Sub

End Class



#If NETCF

' Incomplete and untested !
'
' Because vbc.exe explodes silently (apart from an Event Log message) when told 
' to build for the NETCF...


Class BackgroundWorker

    '-------------------------------------------------------
    Public Sub BackgroundWorker()
    End Sub

    '-------------------------------------------------------
    Public Property WorkerReportsProgress As Boolean
        Get
            ' fake; noop.
        End Get
        Set
            ' fake; noop.
        End Set
    End Property

    '-------------------------------------------------------
    Public Sub RunWorkerAsync(argument As object)
        ' todo
        Dim wcbk As New System.Threading.WaitCallback(AddressOf WorkItem)
        System.Threading.ThreadPool.QueueUserWorkItem(wcbk, argument)
    End Sub

    Private Sub WorkItem(state As Object)
        Dim sender As Object = Nothing
        RaiseEvent DoWork(sender, New DoWorkEventArgs(state))
        RaiseEvent RunWorkerCompleted()
    End Sub

    Public Sub ReportProgress(percentProgress As Int32)
        ' todo
    End Sub

    '-------------------------------------------------------
    Public Event DoWork As DoWorkEventHandler
    Public Event ProgressChanged As ProgressChangedEventHandler
    Public Event RunWorkerCompleted As RunWorkerCompletedEventHandler

End Class


Public Delegate Sub DoWorkEventHandler(ByVal sender As Object, ByVal e As DoWorkEventArgs)

Public Class DoWorkEventArgs 
    Inherits CancelEventArgs
    
    Public Sub New(ByVal argument As Object)
      m_argument = argument
    End Sub
 
    Public Property Result As Object
      Get
            Return m_result
      End Get
      Set(ByVal value As Object)
            m_result = value
      End Set
    End Property

    Public ReadOnly Property Argument As Object
      Get
            Return m_argument
      End Get
    End Property

    ' Fields
    Private m_argument As Object
    Private m_result As Object
End Class


Public Delegate Sub ProgressChangedEventHandler(ByVal sender As Object, ByVal e As ProgressChangedEventArgs)

Public Class ProgressChangedEventArgs
      Inherits EventArgs

      ' Methods
      Public Sub New(ByVal progressPercentage As Integer, ByVal userState As Object)
          m_percentage = progressPercentage
      End Sub

      ' Properties
      Public ReadOnly Property ProgressPercentage As Integer
      Get
            Return m_percentage
      End Get
      End Property

      Public ReadOnly Property UserState As Object
      Get
            Return m_userState
      End Get
      End Property

      ' Fields
      Private ReadOnly m_percentage As Integer
      Private ReadOnly m_userState As Object
End Class


Public Delegate Sub RunWorkerCompletedEventHandler(ByVal sender As Object, ByVal e As RunWorkerCompletedEventArgs)

Public Class RunWorkerCompletedEventArgs
'      Inherits AsyncCompletedEventArgs

      ' Methods
      Public Sub New()'ByVal result As Object, ByVal [error] As Exception, ByVal cancelled As Boolean)
          
      End Sub

      ' Properties
      Public ReadOnly Property Result As Object
      Get
          Return Nothing 'm_result
      End Get
      End Property

      ' Fields
      Private m_result As Object
End Class


#End If


'EOF

